#pragma once
#ifndef _FTL_COMMON_CONFIGURATION_HPP_
#define _FTL_COMMON_CONFIGURATION_HPP_

#include <nlohmann/json.hpp>
//#include <ftl/configurable.hpp>
#include <string>
#include <vector>
#include <optional>

namespace ftl {

class Configurable;

bool is_directory(const std::string &path);
bool is_file(const std::string &path);
bool create_directory(const std::string &path);
bool is_video(const std::string &file);

namespace config {

typedef nlohmann::json json_t;

std::optional<std::string> locateFile(const std::string &name);

Configurable *configure(int argc, char **argv, const std::string &root);

/**
 * Resolve a JSON schema reference, but do not wait for a remote reference
 * if it is not available. A null entity is returned if not resolved.
 */
json_t &resolve(const std::string &);

/**
 * Resolve a reference object, or if not a reference object it simply returns
 * the original object. A reference object with additional properties other
 * than $ref will result in a new merged object.
 */
json_t &resolve(json_t &ref);

/**
 * Resolve a JSON schema reference and block until such a reference can be
 * resolved correctly.
 */
json_t &resolveWait(const std::string &);

/**
 * Using a JSON schema reference, find an existing instance of a Configurable
 * object for that reference. Or return nullptr if not found.
 */
Configurable *find(const std::string &uri);

/**
 * Adds a Configurable instance to the database of instances so that it can
 * then be resolved using find().
 */
void registerConfigurable(Configurable *cfg);

template <typename T, typename... ARGS>
T *create(json_t &link, ARGS ...args);

template <typename T, typename... ARGS>
T *create(ftl::Configurable *parent, const std::string &name, ARGS ...args);

void set(const std::string &uri, const nlohmann::json &);

}  // namespace config

// Deprecated
using config::create;
using config::locateFile;
using config::configure;

}  // namespace ftl

#include <ftl/configurable.hpp>

template <typename T, typename... ARGS>
T *ftl::config::create(json_t &link, ARGS ...args) {
    auto &r = link; // = ftl::config::resolve(link);

    if (!r["$id"].is_string()) {
        LOG(FATAL) << "Entity does not have $id or parent: " << r;
        return nullptr;
    }

    ftl::Configurable *cfg = ftl::config::find(r["$id"].get<std::string>());
    if (!cfg) {
       // try {
            cfg = new T(r, args...);
        //} catch(...) {
       //     LOG(FATAL) << "Could not construct " << link;
        //}
    }

    try {
        return dynamic_cast<T*>(cfg);
    } catch(...) {
        LOG(FATAL) << "Configuration URI object is of wrong type: " << link;
        return nullptr;
    }
}

template <typename T, typename... ARGS>
T *ftl::config::create(ftl::Configurable *parent, const std::string &name, ARGS ...args) {
    nlohmann::json &entity = ftl::config::resolve(parent->getConfig()[name]);

    if (entity.is_object()) {
        if (!entity["$id"].is_string()) {
			// TODO(Nick) Check for # in URI
            entity["$id"] = *parent->get<std::string>("$id") + std::string("/") + name;
        }
    } /*else {
		nlohmann::json &res = resolve(entity);
		if (!res["uri"].is_string()) {
            res["uri"] = *parent->get<std::string>("uri") + std::string("/") + name;
			LOG(WARNING) << "Creating false URI!!! - " << res["uri"].get<std::string>();
        }
	}*/

    return create<T>(entity, args...);
}

#endif  // _FTL_COMMON_CONFIGURATION_HPP_

